import csv
import time
import sys
import traceback
import argparse
import uuid
import datetime
import logging
from collections import defaultdict

from flask import Flask, render_template, redirect, url_for, request, session
import constant, glob
from flask_session import Session
import os, random
from datetime import date
import google_auth_oauthlib.flow
from googleapiclient.discovery import build

from modules.file_utils import FileUtils

GOOGLE_APP_ID = '786124187229-updm7fu1efda9e9usctu5gi74df3hakp.apps.googleusercontent.com'
GOOGLE_SECRET = 'DJa9aFw6FfN_BF6zGK1gVAD8'
GOOGLE_CLIENT_SECRETS_FILE = 'client_secret_786124187229-updm7fu1efda9e9usctu5gi74df3hakp.apps.googleusercontent.com.json'
GOOGLE_SCOPES = [
    'openid',
    'https://www.googleapis.com/auth/userinfo.profile',
    'https://www.googleapis.com/auth/userinfo.email'
]

app = Flask(__name__)
app.config['SESSION_TYPE'] = 'filesystem'
app.config["SESSION_FILE_DIR"] = "sessions"
app.config['SESSION_PERMANENT'] = True
app.config["SESSION_USE_SIGNER"] = True
app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(hours=16)
app.secret_key = "sgZpbXj4Cr7fkdxWt9iPqQnNcHUw3hD625my8YKR"
Session(app)

parser = argparse.ArgumentParser()
parser.add_argument('-port', default='5000', type=str)
parser.add_argument('-is_debug', default=True, type=lambda x: (str(x).lower() == 'true'))
args = parser.parse_known_args()[0]


def error_logging(e):
    logging.exception(str(e))
    exc_type, exc_value, exc_tb = sys.exc_info()
    desc = '\n'.join(traceback.format_exception(exc_type, exc_value, exc_tb)).replace('\\n', '\n')
    desc = desc.replace("\\n", "\n")
    logging.error(desc)


def get_next_sample():
    try:
        test_sample = ""
        columns = defaultdict(list)
        sample_name_list = []
        all_samples = FileUtils.listSubFiles(
            f'{constant.CURRENT_DIR}/{constant.PATH_DIR_WAVS_AUDIO_EXAMPLES}',
            is_include_sub_dirs=True
        )

        all_samples = [it[it.index('static/')+7:].replace('//', '/') for it in all_samples if it.endswith('.wav')]


        total_wav_file_count = len(all_samples)

        if os.path.exists(f'{constant.PATH_DIR_CSV_RESULTS}'):
            with open(f'{constant.PATH_DIR_CSV_RESULTS}') as f:
                reader = csv.DictReader(f)
                for row in reader:
                    for (key, value) in row.items():
                        columns[key].append(value)

            if total_wav_file_count > len(columns['sample']) > 0:
                for sample in all_samples:
                    if sample not in columns['sample']:
                        sample_name_list.append(sample)
                test_sample = random.choice(sample_name_list)

        if test_sample == "":
            test_sample = random.choice(all_samples)

        return test_sample

    except Exception as e:
        error_logging(e)


def get_user_info(credentials):
    try:
        user_info_service = build(
            serviceName='oauth2', version='v2',
            credentials=credentials)
        user_info = None
        try:
            user_info = user_info_service.userinfo().get().execute()
        except Exception as e:
            error_logging(e)

        if user_info and user_info.get('id'):
            user_info = user_info
        else:
            user_info = None

        return user_info
    except Exception as e:
        error_logging(e)

def credentials_to_dict(credentials):
    return {'token': credentials.token,
            'refresh_token': credentials.refresh_token,
            'token_uri': credentials.token_uri,
            'client_id': credentials.client_id,
            'client_secret': credentials.client_secret,
            'scopes': credentials.scopes}


@app.route('/')
def index():
    try:
        session.clear()
        return render_template('index.html')
    except Exception as e:
        error_logging(e)


@app.route('/intro')
def intro():
    try:
        session['counter'] = 0
        session['uid'] = uuid.uuid4()
        session["audio_sample"] = ""
        session['backcounter'] = 0
        if args.is_debug:
            session['credentials'] = {
                'id': 0,
                'email': 'debug'
            }
        if 'credentials' not in session:
            login_state = session['login_state']

            flow = google_auth_oauthlib.flow.Flow.from_client_secrets_file(
                GOOGLE_CLIENT_SECRETS_FILE, scopes=GOOGLE_SCOPES, state=login_state)
            flow.redirect_uri = f'{constant.SITE_URL}/intro'

            authorization_response = request.url
            flow.fetch_token(authorization_response=authorization_response)
            credentials = flow.credentials

            session['credentials'] = get_user_info(credentials)

        return render_template('intro.html', credentials=session['credentials'])
    except Exception as e:
        error_logging(e)


def get_sample_amount_per_session(user_email):
    try:
        switcher = {
            constant.EMAIL_DEBUG: 7,
        }
        return switcher.get(user_email, constant.NUM_SAMPLES_PER_SESSION)
    except Exception as e:
        error_logging(e)


@app.route('/sample', methods=['GET', 'POST'])
def sample():
    try:
        
        if request.method == 'POST':
            
            if request.form.get("back_button") and session['counter'] > 0:
                session['backcounter'] += 1
                if session['backcounter'] == 1:
                    session['counter'] -= 1
                user_progress = session['counter'] / get_sample_amount_per_session(session['credentials']['email'])
                user_progress_percent = int(user_progress * 100)
                return render_template(
                    'screen.html',
                    audio_sample=session['previous'],
                    progress=user_progress_percent,
                    credentials=session['credentials']

                )
            session['backcounter'] = 0
            session['counter'] += 1
            if session['counter'] < get_sample_amount_per_session(session['credentials']['email']):
                has_completed_sampling = False
            else:
                has_completed_sampling = True
                
            problem_sanetize = request.form.get("problem").strip()
            problem_sanetize = problem_sanetize.replace(',', ' ')
            problem_sanetize = problem_sanetize.replace('\n', ' ')

            csv_file_information = {
                'user_id': session['credentials']['id'],
                'session_id': session['uid'],
                'email': session['credentials']['email'],
                'date': date.today().strftime("%d/%m/%Y"),
                'timestamp': time.time(),
                'sample': request.form.get("audio_sample"),
                'counter': session['counter'],
                'time_elapsed': request.form.get("time_elapsed"),
                'label': request.form.get("label"),
                'problem_description': problem_sanetize}

            if os.path.exists(f'{constant.PATH_DIR_CSV_RESULTS}'):
                is_empty = False
            else:
                is_empty = True
            try:
                with open(f'{constant.PATH_DIR_CSV_RESULTS}', mode='a', encoding='UTF-8') as csv_file:
                    FileUtils.lock_file(csv_file)

                    csv_writer = csv.DictWriter(csv_file,
                                                dialect='unix',
                                                fieldnames=csv_file_information.keys())
                    if is_empty:
                        csv_writer.writeheader()

                    csv_writer.writerow(csv_file_information)
                    FileUtils.unlock_file(csv_file)
            except Exception as e:
                error_logging(e)

            if has_completed_sampling:
                return redirect(url_for('end_screen'))
            else:
                return redirect(url_for('sample'))
            
        session['previous'] = session["audio_sample"]
        session["audio_sample"] = get_next_sample()
            
        user_progress = session['counter']/get_sample_amount_per_session(session['credentials']['email'])
        user_progress_percent = int(user_progress * 100)

        return render_template(
            'screen.html',
            audio_sample=session["audio_sample"],
            progress=user_progress_percent,
            credentials=session['credentials']

        )
    except Exception as e:
        error_logging(e)


@app.route('/end')
def end_screen():
    try:
        return render_template('endscreen.html', credentials=session['credentials'])
    except Exception as e:
        error_logging(e)


@app.route('/login')
def login():
    try:
        # Create flow instance to manage the OAuth 2.0 Authorization Grant Flow steps.
        flow = google_auth_oauthlib.flow.Flow.from_client_secrets_file(
            GOOGLE_CLIENT_SECRETS_FILE, scopes=GOOGLE_SCOPES)

        # The URI created here must exactly match one of the authorized redirect URIs
        # for the OAuth 2.0 client, which you configured in the API Console. If this
        # value doesn't match an authorized URI, you will get a 'redirect_uri_mismatch'
        # error.
        flow.redirect_uri = f'{constant.SITE_URL}/intro'

        authorization_url, state = flow.authorization_url(
            # Enable offline access so that you can refresh an access token without
            # re-prompting the user for permission. Recommended for web server apps.
            access_type='offline',
            # Enable incremental authorization. Recommended as a best practice.
            include_granted_scopes='true')

        # Store the state so the callback can verify the auth server response.
        session['login_state'] = state

        return redirect(authorization_url)
    except Exception as e:
        error_logging(e)


if __name__ == '__main__':
    app.run(
        debug=args.is_debug,
        port=args.port
    )
